@@ -82,6 +82,7 @@ struct session_req {
 	guint			id;		/* Listener id */
 	uint8_t			mode;		/* Requested mode */
 	int			refcount;	/* Session refcount */
+	gboolean		got_reply;	/* Agent reply received */
 };
 
 struct service_auth {
@@ -97,6 +98,7 @@ struct btd_adapter {
 	char *path;			/* adapter object path */
 	bdaddr_t bdaddr;		/* adapter Bluetooth Address */
 	guint discov_timeout_id;	/* discoverable timeout id */
+	guint stop_discov_id;		/* stop inquiry/scanning id */
 	uint32_t discov_timeout;	/* discoverable time(sec) */
 	guint pairable_timeout_id;	/* pairable timeout id */
 	uint32_t pairable_timeout;	/* pairable time(sec) */
@@ -201,42 +203,26 @@ static void dev_info_free(struct remote_dev_info *dev)
 	g_free(dev);
 }
 
-void clear_found_devices_list(struct btd_adapter *adapter)
-{
-	if (!adapter->found_devices)
-		return;
-
-	g_slist_foreach(adapter->found_devices, (GFunc) dev_info_free, NULL);
-	g_slist_free(adapter->found_devices);
-	adapter->found_devices = NULL;
-}
-
 static void update_ext_inquiry_response(struct btd_adapter *adapter)
 {
-	uint8_t fec = 0, data[240];
+	uint8_t data[240];
 	struct hci_dev *dev = &adapter->dev;
-	int dd;
+	int ret;
 
 	if (!(dev->features[6] & LMP_EXT_INQ))
 		return;
 
 	memset(data, 0, sizeof(data));
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return;
-
 	if (dev->ssp_mode > 0)
 		create_ext_inquiry_response((char *) dev->name,
 						adapter->tx_power,
 						adapter->services, data);
 
-	if (hci_write_ext_inquiry_response(dd, fec, data,
-						HCI_REQ_TIMEOUT) < 0)
+	ret = adapter_ops->write_eir_data(adapter->dev_id, data);
+	if (ret < 0)
 		error("Can't write extended inquiry response: %s (%d)",
-						strerror(errno), errno);
-
-	hci_close_dev(dd);
+						strerror(-ret), -ret);
 }
 
 static int adapter_set_service_classes(struct btd_adapter *adapter,
@@ -298,10 +284,10 @@ int btd_adapter_set_class(struct btd_adapter *adapter, uint8_t major,
 	return err;
 }
 
-int pending_remote_name_cancel(struct btd_adapter *adapter)
+static int pending_remote_name_cancel(struct btd_adapter *adapter)
 {
 	struct remote_dev_info *dev, match;
-	int err = 0;
+	int err;
 
 	/* find the pending remote name request */
 	memset(&match, 0, sizeof(struct remote_dev_info));
@@ -687,11 +673,17 @@ static uint8_t get_needed_mode(struct btd_adapter *adapter, uint8_t mode)
 	return mode;
 }
 
-static void adapter_stop_inquiry(struct btd_adapter *adapter)
+static void stop_discovery(struct btd_adapter *adapter, gboolean suspend)
 {
 	pending_remote_name_cancel(adapter);
 
-	/* Clear out of range device list */
+	if (suspend == FALSE) {
+		g_slist_foreach(adapter->found_devices,
+				(GFunc) dev_info_free, NULL);
+		g_slist_free(adapter->found_devices);
+		adapter->found_devices = NULL;
+	}
+
 	if (adapter->oor_devices) {
 		g_slist_free(adapter->oor_devices);
 		adapter->oor_devices = NULL;
@@ -699,13 +691,21 @@ static void adapter_stop_inquiry(struct btd_adapter *adapter)
 
 	/* Reset if suspended, otherwise remove timer (software scheduler)
 	   or request inquiry to stop */
-	if (adapter->state & SUSPENDED_INQUIRY)
-		adapter->state &= ~SUSPENDED_INQUIRY;
-	else if (adapter->scheduler_id) {
+	if (adapter->state & STATE_SUSPENDED) {
+		adapter->state &= ~STATE_SUSPENDED;
+		return;
+	}
+
+	if (adapter->scheduler_id) {
 		g_source_remove(adapter->scheduler_id);
 		adapter->scheduler_id = 0;
-	} else
-		adapter_ops->stop_discovery(adapter->dev_id);
+		return;
+	}
+
+	if (adapter->state & STATE_LE_SCAN)
+		adapter_ops->stop_scanning(adapter->dev_id);
+	else
+		adapter_ops->stop_inquiry(adapter->dev_id);
 }
 
 static void session_remove(struct session_req *req)
@@ -742,9 +742,7 @@ static void session_remove(struct session_req *req)
 
 		DBG("Stopping discovery");
 
-		clear_found_devices_list(adapter);
-
-		adapter_stop_inquiry(adapter);
+		stop_discovery(adapter, FALSE);
 	}
 }
 
@@ -757,7 +755,7 @@ static void session_free(struct session_req *req)
 
 	if (req->msg) {
 		dbus_message_unref(req->msg);
-		if (req->mode)
+		if (!req->got_reply && req->mode && req->adapter->agent)
 			agent_cancel(req->adapter->agent);
 	}
 
@@ -794,6 +792,8 @@ static void confirm_mode_cb(struct agent *agent, DBusError *derr, void *data)
 	int err;
 	DBusMessage *reply;
 
+	req->got_reply = TRUE;
+
 	if (derr && dbus_error_is_set(derr)) {
 		reply = dbus_message_new_error(req->msg, derr->name,
 						derr->message);
@@ -961,7 +961,7 @@ void adapter_update_local_name(bdaddr_t *bdaddr, uint8_t status, void *ptr)
 
 	dev = &adapter->dev;
 
-	memcpy(&rp, ptr, MAX_NAME_LENGTH);
+	memcpy(&rp, ptr, sizeof(rp));
 	if (strncmp((char *) rp.name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
 		return;
 
@@ -1031,6 +1031,7 @@ static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
 			return failed_strerror(msg, err);
 
 		adapter->name_stored = TRUE;
+		update_ext_inquiry_response(adapter);
 	}
 
 done:
@@ -1234,20 +1235,49 @@ struct btd_device *adapter_get_device(DBusConnection *conn,
 	return adapter_create_device(conn, adapter, address);
 }
 
-static int adapter_start_inquiry(struct btd_adapter *adapter)
+static gboolean stop_scanning(gpointer user_data)
 {
-	gboolean periodic = TRUE;
+	struct btd_adapter *adapter = user_data;
+
+	adapter_ops->stop_scanning(adapter->dev_id);
+
+	return FALSE;
+}
+
+static int start_discovery(struct btd_adapter *adapter)
+{
+	int err, type;
 
 	/* Do not start if suspended */
-	if (adapter->state & SUSPENDED_INQUIRY)
+	if (adapter->state & STATE_SUSPENDED)
 		return 0;
 
-	if (main_opts.discov_interval)
-		periodic = FALSE;
+	/* Postpone discovery if still resolving names */
+	if (adapter->state & STATE_RESOLVNAME)
+		return 1;
 
 	pending_remote_name_cancel(adapter);
 
-	return adapter_ops->start_discovery(adapter->dev_id, periodic);
+	type = adapter_get_discover_type(adapter) & ~DISC_RESOLVNAME;
+
+	switch (type) {
+	case DISC_STDINQ:
+	case DISC_INTERLEAVE:
+		err = adapter_ops->start_inquiry(adapter->dev_id,
+							0x08, FALSE);
+		break;
+	case DISC_PINQ:
+		err = adapter_ops->start_inquiry(adapter->dev_id,
+							0x08, TRUE);
+		break;
+	case DISC_LE:
+		err = adapter_ops->start_scanning(adapter->dev_id);
+		break;
+	default:
+		err = -1;
+	}
+
+	return err;
 }
 
 static DBusMessage *adapter_start_discovery(DBusConnection *conn,
@@ -1270,10 +1300,7 @@ static DBusMessage *adapter_start_discovery(DBusConnection *conn,
 	if (adapter->disc_sessions)
 		goto done;
 
-	if (main_opts.name_resolv)
-		adapter->state |= RESOLVE_NAME;
-
-	err = adapter_start_inquiry(adapter);
+	err = start_discovery(adapter);
 	if (err < 0)
 		return failed_strerror(msg, -err);
 
@@ -1378,7 +1405,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 				DBUS_TYPE_UINT32, &adapter->pairable_timeout);
 
 
-	if (adapter->state & PERIODIC_INQUIRY || adapter->state & STD_INQUIRY)
+	if (adapter->state & (STATE_PINQ | STATE_STDINQ | STATE_LE_SCAN))
 		value = TRUE;
 	else
 		value = FALSE;
@@ -1913,48 +1940,15 @@ static inline uint8_t get_inquiry_mode(struct hci_dev *dev)
 	return 0;
 }
 
-static int adapter_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
-{
-	int dd, err;
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0) {
-		err = -errno;
-		error("Can't open device hci%d: %s (%d)",
-					dev_id, strerror(errno), errno);
-		return err;
-	}
-
-	if (hci_read_bd_addr(dd, bdaddr, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
-		error("Can't read address for hci%d: %s (%d)",
-					dev_id, strerror(errno), errno);
-		hci_close_dev(dd);
-		return err;
-	}
-
-	hci_close_dev(dd);
-
-	return 0;
-}
-
 static int adapter_setup(struct btd_adapter *adapter, const char *mode)
 {
 	struct hci_dev *dev = &adapter->dev;
 	uint8_t events[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0x00, 0x00 };
 	uint8_t inqmode;
-	int err , dd;
+	int err;
 	char name[MAX_NAME_LENGTH + 1];
 	uint8_t cls[3];
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		err = -errno;
-		error("Can't open device hci%d: %s (%d)", adapter->dev_id,
-						strerror(errno), errno);
-		return err;
-	}
-
 	if (dev->lmp_ver > 1) {
 		if (dev->features[5] & LMP_SNIFF_SUBR)
 			events[5] |= 0x20;
@@ -1984,25 +1978,35 @@ static int adapter_setup(struct btd_adapter *adapter, const char *mode)
 						 * Features Notification */
 		}
 
-		hci_send_cmd(dd, OGF_HOST_CTL, OCF_SET_EVENT_MASK,
-						sizeof(events), events);
+		if (dev->features[4] & LMP_LE)
+			events[7] |= 0x20;	/* LE Meta-Event */
+
+		adapter_ops->set_event_mask(adapter->dev_id, events,
+							sizeof(events));
 	}
 
 	inqmode = get_inquiry_mode(dev);
 	if (inqmode < 1)
-		goto done;
+		return 0;
 
-	if (hci_write_inquiry_mode(dd, inqmode, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
+	err = adapter_ops->write_inq_mode(adapter->dev_id, inqmode);
+	if (err < 0) {
 		error("Can't write inquiry mode for %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		hci_close_dev(dd);
+					adapter->path, strerror(-err), -err);
 		return err;
 	}
 
 	if (dev->features[7] & LMP_INQ_TX_PWR)
-		hci_send_cmd(dd, OGF_HOST_CTL,
-				OCF_READ_INQ_RESPONSE_TX_POWER_LEVEL, 0, NULL);
+		adapter_ops->read_inq_tx_pwr(adapter->dev_id);
+
+	if (dev->features[4] & LMP_LE) {
+		err = adapter_ops->write_le_host(adapter->dev_id, 0x01, 0x00);
+		if (err < 0) {
+			error("Can't write LE host supported for %s: %s(%d)",
+					adapter->path, strerror(-err), -err);
+			return err;
+		}
+	}
 
 	if (read_local_name(&adapter->bdaddr, name) < 0)
 		expand_name(name, MAX_NAME_LENGTH, main_opts.name,
@@ -2021,12 +2025,11 @@ static int adapter_setup(struct btd_adapter *adapter, const char *mode)
 		if (class)
 			memcpy(cls, &class, 3);
 		else
-			goto done;
+			return 0;
 	}
 
 	btd_adapter_set_class(adapter, cls[1], cls[0]);
-done:
-	hci_close_dev(dd);
+
 	return 0;
 }
 
@@ -2107,21 +2110,24 @@ static void load_devices(struct btd_adapter *adapter)
 	textfile_foreach(filename, create_stored_device_from_blocked, adapter);
 }
 
+int btd_adapter_block_address(struct btd_adapter *adapter, bdaddr_t *bdaddr)
+{
+	return adapter_ops->block_device(adapter->dev_id, bdaddr);
+}
+
+int btd_adapter_unblock_address(struct btd_adapter *adapter, bdaddr_t *bdaddr)
+{
+	return adapter_ops->unblock_device(adapter->dev_id, bdaddr);
+}
+
 static void clear_blocked(struct btd_adapter *adapter)
 {
-	int dd;
+	int err;
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		error("hci_open_dev(hci%d): %s (%d)", adapter->dev_id,
-						strerror(errno), errno);
-		return;
-	}
-
-	if (ioctl(dd, HCIUNBLOCKADDR, BDADDR_ANY) < 0)
-		error("ioctl(HCIUNBLOCKADDR): %s (%d)", strerror(errno), errno);
-
-	hci_close_dev(dd);
+	err = adapter_ops->unblock_device(adapter->dev_id, BDADDR_ANY);
+	if (err < 0)
+		error("Clearing blocked list failed: %s (%d)",
+						strerror(-err), -err);
 }
 
 static void probe_driver(gpointer data, gpointer user_data)
@@ -2154,27 +2160,18 @@ static void load_drivers(struct btd_adapter *adapter)
 
 static void load_connections(struct btd_adapter *adapter)
 {
-	struct hci_conn_list_req *cl = NULL;
-	struct hci_conn_info *ci;
-	int i, dd;
+	GSList *l, *conns;
+	int err;
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return;
-
-	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
-
-	cl->dev_id = adapter->dev_id;
-	cl->conn_num = 10;
-	ci = cl->conn_info;
-
-	if (ioctl(dd, HCIGETCONNLIST, cl) != 0) {
-		g_free(cl);
-		hci_close_dev(dd);
+	err = adapter_ops->get_conn_list(adapter->dev_id, &conns);
+	if (err < 0) {
+		error("Unable to fetch existing connections: %s (%d)",
+							strerror(-err), -err);
 		return;
 	}
 
-	for (i = 0; i < cl->conn_num; i++, ci++) {
+	for (l = conns; l != NULL; l = g_slist_next(l)) {
+		struct hci_conn_info *ci = l->data;
 		struct btd_device *device;
 		char address[18];
 
@@ -2184,8 +2181,8 @@ static void load_connections(struct btd_adapter *adapter)
 			adapter_add_connection(adapter, device, ci->handle);
 	}
 
-	g_free(cl);
-	hci_close_dev(dd);
+	g_slist_foreach(conns, (GFunc) g_free, NULL);
+	g_slist_free(conns);
 }
 
 static int get_discoverable_timeout(const char *src)
@@ -2245,6 +2242,66 @@ static void call_adapter_powered_callbacks(struct btd_adapter *adapter,
        }
 }
 
+static void emit_device_disappeared(gpointer data, gpointer user_data)
+{
+	struct remote_dev_info *dev = data;
+	struct btd_adapter *adapter = user_data;
+	char address[18];
+	const char *paddr = address;
+
+	ba2str(&dev->bdaddr, address);
+
+	g_dbus_emit_signal(connection, adapter->path,
+			ADAPTER_INTERFACE, "DeviceDisappeared",
+			DBUS_TYPE_STRING, &paddr,
+			DBUS_TYPE_INVALID);
+
+	adapter->found_devices = g_slist_remove(adapter->found_devices, dev);
+}
+
+static void update_oor_devices(struct btd_adapter *adapter)
+{
+	g_slist_foreach(adapter->oor_devices, emit_device_disappeared, adapter);
+	g_slist_foreach(adapter->oor_devices, (GFunc) dev_info_free, NULL);
+	g_slist_free(adapter->oor_devices);
+	adapter->oor_devices =  g_slist_copy(adapter->found_devices);
+}
+
+static gboolean bredr_capable(struct btd_adapter *adapter)
+{
+	struct hci_dev *dev = &adapter->dev;
+
+	return (dev->features[4] & LMP_NO_BREDR) == 0 ? TRUE : FALSE;
+}
+
+static gboolean le_capable(struct btd_adapter *adapter)
+{
+	struct hci_dev *dev = &adapter->dev;
+
+	return (dev->features[4] & LMP_LE &&
+			dev->extfeatures[0] & LMP_HOST_LE) ? TRUE : FALSE;
+}
+
+int adapter_get_discover_type(struct btd_adapter *adapter)
+{
+	gboolean le, bredr;
+	int type;
+
+	le = le_capable(adapter);
+	bredr = bredr_capable(adapter);
+
+	if (le)
+		type = bredr ? DISC_INTERLEAVE : DISC_LE;
+	else
+		type = main_opts.discov_interval ? DISC_STDINQ :
+							DISC_PINQ;
+
+	if (main_opts.name_resolv)
+		type |= DISC_RESOLVNAME;
+
+	return type;
+}
+
 static int adapter_up(struct btd_adapter *adapter, const char *mode)
 {
 	char srcaddr[18];
@@ -2258,7 +2315,7 @@ static int adapter_up(struct btd_adapter *adapter, const char *mode)
 	adapter->up = 1;
 	adapter->discov_timeout = get_discoverable_timeout(srcaddr);
 	adapter->pairable_timeout = get_pairable_timeout(srcaddr);
-	adapter->state = DISCOVER_TYPE_NONE;
+	adapter->state = STATE_IDLE;
 	adapter->mode = MODE_CONNECTABLE;
 	adapter->cache_enable = TRUE;
 	scan_mode = SCAN_PAGE;
@@ -2338,7 +2395,7 @@ int adapter_start(struct btd_adapter *adapter)
 	struct hci_dev_info di;
 	struct hci_version ver;
 	uint8_t features[8];
-	int dd, err;
+	int err;
 	char mode[14], address[18];
 
 	if (hci_devinfo(adapter->dev_id, &di) < 0)
@@ -2354,7 +2411,7 @@ int adapter_start(struct btd_adapter *adapter)
 
 		DBG("Adapter %s without an address", adapter->path);
 
-		err = adapter_read_bdaddr(adapter->dev_id, &di.bdaddr);
+		err = adapter_ops->read_bdaddr(adapter->dev_id, &di.bdaddr);
 		if (err < 0)
 			return err;
 	}
@@ -2372,19 +2429,10 @@ int adapter_start(struct btd_adapter *adapter)
 			strcpy(mode, "connectable");
 	}
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		err = -errno;
-		error("Can't open adapter %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		return err;
-	}
-
-	if (hci_read_local_version(dd, &ver, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
+	err = adapter_ops->read_local_version(adapter->dev_id, &ver);
+	if (err < 0) {
 		error("Can't read version info for %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		hci_close_dev(dd);
+					adapter->path, strerror(-err), -err);
 		return err;
 	}
 
@@ -2393,11 +2441,10 @@ int adapter_start(struct btd_adapter *adapter)
 	dev->lmp_subver = ver.lmp_subver;
 	dev->manufacturer = ver.manufacturer;
 
-	if (hci_read_local_features(dd, features, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
+	err = adapter_ops->read_local_features(adapter->dev_id, features);
+	if (err < 0) {
 		error("Can't read features for %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		hci_close_dev(dd);
+					adapter->path, strerror(-err), -err);
 		return err;
 	}
 
@@ -2406,22 +2453,10 @@ int adapter_start(struct btd_adapter *adapter)
 	if (!(features[6] & LMP_SIMPLE_PAIR))
 		goto setup;
 
-	if (ioctl(dd, HCIGETAUTHINFO, NULL) < 0 && errno != EINVAL)
-		hci_write_simple_pairing_mode(dd, 0x01, HCI_REQ_TIMEOUT);
-
-	if (hci_read_simple_pairing_mode(dd, &dev->ssp_mode,
-						HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
-		error("Can't read simple pairing mode on %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		/* Fall through since some chips have broken
-		 * read_simple_pairing_mode behavior */
-	}
+	adapter_ops->init_ssp_mode(adapter->dev_id, &dev->ssp_mode);
 
 setup:
-	hci_send_cmd(dd, OGF_LINK_POLICY, OCF_READ_DEFAULT_LINK_POLICY,
-								0, NULL);
-	hci_close_dev(dd);
+	adapter_ops->read_link_policy(adapter->dev_id);
 
 	adapter->current_cod = 0;
 
@@ -2429,7 +2464,7 @@ setup:
 
 	if (!adapter->initialized && adapter->already_up) {
 		DBG("Stopping Inquiry at adapter startup");
-		adapter_ops->stop_discovery(adapter->dev_id);
+		adapter_ops->stop_inquiry(adapter->dev_id);
 	}
 
 	err = adapter_up(adapter, mode);
@@ -2454,12 +2489,6 @@ static void reply_pending_requests(struct btd_adapter *adapter)
 			device_cancel_bonding(device,
 						HCI_OE_USER_ENDED_CONNECTION);
 	}
-
-	if (adapter->state & STD_INQUIRY || adapter->state & PERIODIC_INQUIRY) {
-		/* Cancel inquiry initiated by D-Bus client */
-		if (adapter->disc_sessions)
-			adapter_ops->stop_discovery(adapter->dev_id);
-	}
 }
 
 static void unload_drivers(struct btd_adapter *adapter)
@@ -2487,6 +2516,8 @@ int adapter_stop(struct btd_adapter *adapter)
 	/* check pending requests */
 	reply_pending_requests(adapter);
 
+	stop_discovery(adapter, FALSE);
+
 	if (adapter->disc_sessions) {
 		g_slist_foreach(adapter->disc_sessions, (GFunc) session_free,
 				NULL);
@@ -2494,13 +2525,6 @@ int adapter_stop(struct btd_adapter *adapter)
 		adapter->disc_sessions = NULL;
 	}
 
-	clear_found_devices_list(adapter);
-
-	if (adapter->oor_devices) {
-		g_slist_free(adapter->oor_devices);
-		adapter->oor_devices = NULL;
-	}
-
 	while (adapter->connections) {
 		struct btd_device *device = adapter->connections->data;
 		adapter_remove_connection(adapter, device, 0);
@@ -2527,7 +2551,7 @@ int adapter_stop(struct btd_adapter *adapter)
 	adapter->up = 0;
 	adapter->scan_mode = SCAN_DISABLED;
 	adapter->mode = MODE_OFF;
-	adapter->state = DISCOVER_TYPE_NONE;
+	adapter->state = STATE_IDLE;
 	adapter->cache_enable = TRUE;
 	adapter->pending_cod = 0;
 	adapter->off_requested = FALSE;
@@ -2625,8 +2649,6 @@ struct btd_adapter *adapter_create(DBusConnection *conn, int id,
 	}
 
 	adapter->dev_id = id;
-	if (main_opts.name_resolv)
-		adapter->state |= RESOLVE_NAME;
 	adapter->path = g_strdup(path);
 	adapter->already_up = devup;
 
@@ -2683,29 +2705,75 @@ void adapter_get_address(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 
 void adapter_set_state(struct btd_adapter *adapter, int state)
 {
-	gboolean discov_active = FALSE;
 	const char *path = adapter->path;
+	gboolean discov_active;
+	int previous, type;
 
 	if (adapter->state == state)
 		return;
 
-	if (state & PERIODIC_INQUIRY || state & STD_INQUIRY)
-		discov_active = TRUE;
-	else if (adapter->disc_sessions && main_opts.discov_interval)
-		adapter->scheduler_id = g_timeout_add_seconds(
-						main_opts.discov_interval,
-						(GSourceFunc) adapter_start_inquiry,
-						adapter);
+	previous = adapter->state;
+	adapter->state = state;
 
-	/* Send out of range */
-	if (!discov_active)
-		adapter_update_oor_devices(adapter);
+	type = adapter_get_discover_type(adapter);
+
+	switch (state) {
+	case STATE_STDINQ:
+	case STATE_PINQ:
+		discov_active = TRUE;
+
+		/* Started a new session while resolving names ? */
+		if (previous & STATE_RESOLVNAME)
+			return;
+		break;
+	case STATE_LE_SCAN:
+		/* Scanning enabled */
+		adapter->stop_discov_id = g_timeout_add(5120,
+						stop_scanning, adapter);
+
+		/* For dual mode: don't send "Discovering = TRUE"  */
+		if (bredr_capable(adapter) == TRUE)
+			return;
+
+		/* LE only */
+		discov_active = TRUE;
+
+		break;
+	case STATE_IDLE:
+		/*
+		 * Interleave: from inquiry to scanning. Interleave is not
+		 * applicable to requests triggered by external applications.
+		 */
+		if (adapter->disc_sessions && (type & DISC_INTERLEAVE) &&
+						(previous & STATE_STDINQ)) {
+			adapter_ops->start_scanning(adapter->dev_id);
+			return;
+		}
+		/* BR/EDR only: inquiry finished */
+		discov_active = FALSE;
+		break;
+	default:
+		discov_active = FALSE;
+		break;
+	}
+
+	if (discov_active == FALSE) {
+		update_oor_devices(adapter);
+		if (type & DISC_RESOLVNAME) {
+			if (adapter_resolve_names(adapter) == 0) {
+				adapter->state |= STATE_RESOLVNAME;
+				return;
+			}
+		}
+	} else if (adapter->disc_sessions && main_opts.discov_interval)
+			adapter->scheduler_id = g_timeout_add_seconds(
+						main_opts.discov_interval,
+						(GSourceFunc) start_discovery,
+						adapter);
 
 	emit_property_changed(connection, path,
 				ADAPTER_INTERFACE, "Discovering",
 				DBUS_TYPE_BOOLEAN, &discov_active);
-
-	adapter->state = state;
 }
 
 int adapter_get_state(struct btd_adapter *adapter)
@@ -3002,33 +3070,6 @@ int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 	return 0;
 }
 
-void adapter_update_oor_devices(struct btd_adapter *adapter)
-{
-	GSList *l;
-
-	for (l = adapter->oor_devices; l; l = l->next) {
-		char address[18];
-		const char *paddr = address;
-		struct remote_dev_info *dev = l->data;
-
-		ba2str(&dev->bdaddr, address);
-
-		g_dbus_emit_signal(connection, adapter->path,
-				ADAPTER_INTERFACE, "DeviceDisappeared",
-				DBUS_TYPE_STRING, &paddr,
-				DBUS_TYPE_INVALID);
-
-		adapter->found_devices = g_slist_remove(adapter->found_devices,
-							dev);
-		dev_info_free(dev);
-	}
-
-	g_slist_free(adapter->oor_devices);
-	adapter->oor_devices = NULL;
-
-	adapter->oor_devices = g_slist_copy(adapter->found_devices);
-}
-
 static void set_mode_complete(struct btd_adapter *adapter)
 {
 	struct session_req *pending;
@@ -3192,13 +3233,13 @@ gboolean adapter_has_discov_sessions(struct btd_adapter *adapter)
 void adapter_suspend_discovery(struct btd_adapter *adapter)
 {
 	if (adapter->disc_sessions == NULL ||
-			adapter->state & SUSPENDED_INQUIRY)
+			adapter->state & STATE_SUSPENDED)
 		return;
 
 	DBG("Suspending discovery");
 
-	adapter_stop_inquiry(adapter);
-	adapter->state |= SUSPENDED_INQUIRY;
+	stop_discovery(adapter, TRUE);
+	adapter->state |= STATE_SUSPENDED;
 }
 
 void adapter_resume_discovery(struct btd_adapter *adapter)
@@ -3208,8 +3249,8 @@ void adapter_resume_discovery(struct btd_adapter *adapter)
 
 	DBG("Resuming discovery");
 
-	adapter->state &= ~SUSPENDED_INQUIRY;
-	adapter_start_inquiry(adapter);
+	adapter->state &= ~STATE_SUSPENDED;
+	start_discovery(adapter);
 }
 
 int btd_register_adapter_driver(struct btd_adapter_driver *driver)
@@ -3519,3 +3560,92 @@ int btd_adapter_set_fast_connectable(struct btd_adapter *adapter,
 
 	return adapter_ops->set_fast_connectable(adapter->dev_id, enable);
 }
+
+int btd_adapter_read_clock(struct btd_adapter *adapter, int handle, int which,
+			int timeout, uint32_t *clock, uint16_t *accuracy)
+{
+	if (!adapter_ops)
+		return -EINVAL;
+
+	if (!adapter->up)
+		return -EINVAL;
+
+	return adapter_ops->read_clock(adapter->dev_id, handle, which,
+						timeout, clock, accuracy);
+}
+
+int btd_adapter_get_conn_handle(struct btd_adapter *adapter,
+				const bdaddr_t *bdaddr, int *handle)
+{
+	if (!adapter_ops)
+		return -EINVAL;
+
+	if (!adapter->up)
+		return -EINVAL;
+
+	return adapter_ops->get_conn_handle(adapter->dev_id, bdaddr, handle);
+}
+
+int btd_adapter_disconnect_device(struct btd_adapter *adapter, uint16_t handle)
+{
+	return adapter_ops->disconnect(adapter->dev_id, handle);
+}
+
+int btd_adapter_remove_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr)
+{
+	return adapter_ops->remove_bonding(adapter->dev_id, bdaddr);
+}
+
+int btd_adapter_request_authentication(struct btd_adapter *adapter,
+					uint16_t handle, uint8_t *status)
+{
+	return adapter_ops->request_authentication(adapter->dev_id,
+							handle, status);
+}
+
+int btd_adapter_pincode_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+							const char *pin)
+{
+	return adapter_ops->pincode_reply(adapter->dev_id, bdaddr, pin);
+}
+
+int btd_adapter_confirm_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+							gboolean success)
+{
+	return adapter_ops->confirm_reply(adapter->dev_id, bdaddr, success);
+}
+
+int btd_adapter_passkey_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+							uint32_t passkey)
+{
+	return adapter_ops->passkey_reply(adapter->dev_id, bdaddr, passkey);
+}
+
+int btd_adapter_get_auth_info(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+								uint8_t *auth)
+{
+	return adapter_ops->get_auth_info(adapter->dev_id, bdaddr, auth);
+}
+
+int btd_adapter_read_scan_enable(struct btd_adapter *adapter)
+{
+	return adapter_ops->read_scan_enable(adapter->dev_id);
+}
+
+int btd_adapter_read_ssp_mode(struct btd_adapter *adapter)
+{
+	return adapter_ops->read_ssp_mode(adapter->dev_id);
+}
+
+int btd_adapter_read_local_ext_features(struct btd_adapter *adapter)
+{
+	return adapter_ops->read_local_ext_features(adapter->dev_id);
+}
+
+void btd_adapter_update_local_ext_features(struct btd_adapter *adapter,
+						const uint8_t *features)
+{
+	struct hci_dev *dev = &adapter->dev;
+
+	memcpy(dev->extfeatures, features, 8);
+}
