@@ -31,14 +31,20 @@
 
 #define ADAPTER_INTERFACE	"org.bluez.Adapter"
 
-/* Discover types */
-#define DISCOVER_TYPE_NONE	0x00
-#define STD_INQUIRY		0x01
-#define PERIODIC_INQUIRY	0x02
-#define SUSPENDED_INQUIRY	0x04
+/* Discover states */
+#define STATE_IDLE		0x00
+#define STATE_LE_SCAN		0x01
+#define STATE_STDINQ		0x02
+#define STATE_PINQ		0x04
+#define STATE_RESOLVNAME	0x08
+#define STATE_SUSPENDED		0x10
 
-/* Actions executed after inquiry complete */
-#define RESOLVE_NAME		0x10
+/* Supported host/controller discover type */
+#define DISC_LE			0x01
+#define DISC_STDINQ		0x02
+#define DISC_INTERLEAVE		0x04
+#define DISC_PINQ		0x08
+#define DISC_RESOLVNAME		0x10
 
 #define MAX_NAME_LENGTH		248
 
@@ -69,6 +75,7 @@ struct hci_dev {
 	int ignore;
 
 	uint8_t  features[8];
+	uint8_t  extfeatures[8];
 	uint8_t  lmp_ver;
 	uint16_t lmp_subver;
 	uint16_t hci_rev;
@@ -99,12 +106,8 @@ void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
 struct btd_device *adapter_create_device(DBusConnection *conn,
 				struct btd_adapter *adapter, const char *address);
 
-int pending_remote_name_cancel(struct btd_adapter *adapter);
-
 int adapter_resolve_names(struct btd_adapter *adapter);
 
-void clear_found_devices_list(struct btd_adapter *adapter);
-
 struct btd_adapter *adapter_create(DBusConnection *conn, int id,
 				gboolean devup);
 void adapter_remove(struct btd_adapter *adapter);
@@ -113,6 +116,7 @@ const gchar *adapter_get_path(struct btd_adapter *adapter);
 void adapter_get_address(struct btd_adapter *adapter, bdaddr_t *bdaddr);
 void adapter_set_state(struct btd_adapter *adapter, int state);
 int adapter_get_state(struct btd_adapter *adapter);
+int adapter_get_discover_type(struct btd_adapter *adapter);
 gboolean adapter_is_ready(struct btd_adapter *adapter);
 struct remote_dev_info *adapter_search_found_devices(struct btd_adapter *adapter,
 						struct remote_dev_info *match);
@@ -123,7 +127,6 @@ void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr);
 void adapter_emit_device_found(struct btd_adapter *adapter,
 				struct remote_dev_info *dev, uint8_t *eir_data);
-void adapter_update_oor_devices(struct btd_adapter *adapter);
 void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode);
 void adapter_setname_complete(bdaddr_t *local, uint8_t status);
 void adapter_update_tx_power(bdaddr_t *bdaddr, uint8_t status, void *ptr);
@@ -184,8 +187,11 @@ struct btd_adapter_ops {
 	int (*set_discoverable) (int index);
 	int (*set_limited_discoverable) (int index, uint32_t class,
 						gboolean limited);
-	int (*start_discovery) (int index, gboolean periodic);
-	int (*stop_discovery) (int index);
+	int (*start_inquiry) (int index, uint8_t length, gboolean periodic);
+	int (*stop_inquiry) (int index);
+	int (*start_scanning) (int index);
+	int (*stop_scanning) (int index);
+
 	int (*resolve_name) (int index, bdaddr_t *bdaddr);
 	int (*cancel_resolve_name) (int index, bdaddr_t *bdaddr);
 	int (*set_name) (int index, const char *name);
@@ -205,6 +211,7 @@ struct btd_adapter_ops {
 	int (*get_conn_list) (int index, GSList **conns);
 	int (*read_local_version) (int index, struct hci_version *ver);
 	int (*read_local_features) (int index, uint8_t *features);
+	int (*read_local_ext_features) (int index);
 	int (*init_ssp_mode) (int index, uint8_t *ssp_mode);
 	int (*read_link_policy) (int index);
 	int (*disconnect) (int index, uint16_t handle);
@@ -215,6 +222,9 @@ struct btd_adapter_ops {
 	int (*confirm_reply) (int index, bdaddr_t *bdaddr, gboolean success);
 	int (*passkey_reply) (int index, bdaddr_t *bdaddr, uint32_t passkey);
 	int (*get_auth_info) (int index, bdaddr_t *bdaddr, uint8_t *auth);
+	int (*read_scan_enable) (int index);
+	int (*read_ssp_mode) (int index);
+	int (*write_le_host) (int index, uint8_t le, uint8_t simul);
 };
 
 int btd_register_adapter_ops(struct btd_adapter_ops *btd_adapter_ops);
@@ -260,3 +270,10 @@ int btd_adapter_passkey_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 
 int btd_adapter_get_auth_info(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 								uint8_t *auth);
+
+int btd_adapter_read_scan_enable(struct btd_adapter *adapter);
+
+int btd_adapter_read_ssp_mode(struct btd_adapter *adapter);
+int btd_adapter_read_local_ext_features(struct btd_adapter *adapter);
+void btd_adapter_update_local_ext_features(struct btd_adapter *adapter,
+						const uint8_t *features);
